/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 1997,2005          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/* AIX Parallel Environment Sample Program             */
/* 5765-D93 (C) Copyright IBM Corporation 1997,1999    */
/* Licensed Materials - Property of IBM                */
/* All rights reserved.                                */
/* US Government Users Restricted Rights -             */
/* Use, duplication or disclosure restricted by        */
/* GSA ADP Schedule Contract with IBM Corp.            */

 /*******************************************************
 * Threaded "ring" program: each task sends a message   *
 * to the next higher task number (mod N), and receives *
 * a message from the next lower task number (mod N).   *
 * The sends and receives are performed on separate     *
 * POSIX threads.                                       *
 * An MPI Barrier on the main thread ensures completion *
 * of all sends and receives.                           *
 *******************************************************/

#include <pthread.h>
#include <stdio.h>
#include <mpi.h>

/* Size of message to be sent, in words */
#define ARRAY_SIZE 10000
/* contents of message - each word gets this value */
#define VALUE 123
/* message tag to use */
#define TAG 5

/* global variables - message buffers */
int in[ARRAY_SIZE],out[ARRAY_SIZE];
/* number of MPI tasks in job, rank of current task */
int tasks,me;
 
int main()
{ 
   int i;

/* thread data */
   pthread_t send_tid, recv_tid;
#ifndef LINUX
   pthread_attr_t ring_attr;
#endif
/* thread functions */
   void *send_thread(void*);
   void *recv_thread(void*);

/* Initialize the thread attributes */
#ifndef LINUX
   pthread_attr_init(&ring_attr);
   pthread_attr_setdetachstate(&ring_attr,PTHREAD_CREATE_UNDETACHED);
#endif

/* Initialize the data array */
   for (i=0;i<ARRAY_SIZE;i++) out[i]=VALUE;
 
/* Initialize MPI */
   MPI_Init(0,NULL);
   MPI_Comm_size(MPI_COMM_WORLD,&tasks);
   MPI_Comm_rank(MPI_COMM_WORLD,&me);
 
/* Create a send thread and a receive thread */
#ifndef LINUX
   pthread_create(&send_tid, &ring_attr, send_thread, NULL);
   pthread_create(&recv_tid, &ring_attr, recv_thread, NULL);
#else
   pthread_create(&send_tid, NULL, send_thread, NULL);
   pthread_create(&recv_tid, NULL, recv_thread, NULL);
#endif

/* Join with the threads after they have done their work */
   pthread_join(send_tid,NULL);
   pthread_join(recv_tid,NULL);	

   MPI_Barrier(MPI_COMM_WORLD);

/* Test the result */
   for (i=0; i< ARRAY_SIZE; i++) {
   if(in[i] != VALUE )
      fprintf(stderr,"ERROR in task %d: in[%d] = %d\n",me,i,in[i]);
	  }

/* Terminate MPI */
   MPI_Finalize(); 

   if(me == 0)  fprintf(stderr,"TEST COMPLETE\n");
   exit(0);
}

/* The sending thread */
void* send_thread(void* dummy)
{
   int dest;
   int rc;
   MPI_Status status;
   MPI_Request msgid;

/* determine the destination task number */
   dest = (me==tasks-1) ? 0 : me+1; 
   rc = MPI_Isend(out,ARRAY_SIZE,MPI_INT,dest,TAG,MPI_COMM_WORLD,&msgid);
   if (rc != MPI_SUCCESS) {
		fprintf(stderr,"error return code = %d from MPI_Isend\n",rc);
        exit(rc);
		}
   rc = MPI_Waitall(1,&msgid,&status);
   if (rc != MPI_SUCCESS) {
		fprintf(stderr,"error return code = %d from MPI_Waitall\n",rc);
        exit(rc);
		}
   return;
	}

/* The receiving thread */
void* recv_thread(void* dummy)
{
   int src;
   int rc;
   MPI_Status status;

/* determine the destination task number */
   src = (me==0) ? tasks-1 : me-1; 
   rc = MPI_Recv(in,ARRAY_SIZE,MPI_INT,src,TAG,MPI_COMM_WORLD,&status);
   if (rc != MPI_SUCCESS) {
		fprintf(stderr,"error return code = %d from MPI_Recv\n",rc);
        exit(rc);
		}
   return;
	}
